#!/usr/bin/env python3

import sys
import os
import pytest
import tempfile
import zipfile
from pathlib import Path

from osgeo import gdal
from owslib.fes import PropertyIsEqualTo

sys.path.insert(0, str(Path(__file__).parent.parent))
from st_lucas import LucasIO, LucasRequest, __version__
from st_lucas.exceptions import LucasDownloadError

class TestST_LUCAS:
    num_of_features = 1
    point_id = 28382290
    
    def _request(self, st=True):
        request = LucasRequest()
        request.operator=PropertyIsEqualTo
        request.propertyname = 'point_id'
        request.literal = str(self.point_id)
        request.st_aggregated = st

        return request

    def _download(self, st=True):
        request = self._request(st)
        lucasio = LucasIO()
        lucasio.download(request)
        return lucasio
    
    def test_001(self):
        """Build a request.

        This tests case consists of checking that LucasRequest.build()
        returns request based on specified filters.
        """
        data = self._request().build()
        
        assert data['typename'] == 'lucas:lucas_st_points'
        assert data['filter'] == f'<ogc:PropertyIsEqualTo xmlns:ogc="http://www.opengis.net/ogc"><ogc:PropertyName>point_id</ogc:PropertyName><ogc:Literal>{self.point_id}</ogc:Literal></ogc:PropertyIsEqualTo>'
    
    def test_002(self):
        """Download LUCAS subset based on request.

        This tests case consists of checking that LucasIO.download()
        retrieves expected LUCAS subset from remote server.
        """
        assert self._download().count() == self.num_of_features

    def test_003(self):
        """Identify LUCAS metadata.

        This tests case consists of checking that LucasIO.metadata
        returns expected LUCAS metadata directory.
        """
        md = self._download().metadata
        
        assert md["LUCAS_TABLE"] == "lucas_st_points"
        assert md["LUCAS_ST"] == "1"
        assert md["LUCAS_CLIENT_VERSION"] == str(__version__)
        assert float(md["LUCAS_DB_VERSION"]) >= 0.9
        assert int(md["LUCAS_MAX_FEATURES"]) > 0

    def test_004(self):
        """Save LUCAS subset to GeoPackage format.

        This tests case consists of checking that created GeoPackage file
        can be open by GDAL library and contains expected number of
        features.
        """
        gpkg_file = Path(tempfile.gettempdir()) / Path(str(os.getpid()) + '.gpkg')
        self._download().to_gpkg(gpkg_file)
        ds = gdal.OpenEx(str(gpkg_file), gdal.OF_VECTOR | gdal.OF_READONLY)
        assert ds.GetLayer().GetFeatureCount() == self.num_of_features
        del ds

    def test_005(self):
        """Identify photos for LUCAS point/year.

        This tests case consists of checking that LucasIO.get_images()
        returns expected directory of images.
        """
        lucasio = self._download(st=False)
        images = lucasio.get_images(2018, self.point_id)

        assert len(images.keys()) == 5

        # download photos
        with tempfile.TemporaryDirectory() as tmpdirname:
            zip_file = lucasio.download_images(images, tmpdirname)

            with zipfile.ZipFile(zip_file) as zp:
                file_list = zp.namelist()

                assert len(file_list) == 5
                for file_name in file_list:
                    assert file_name.endswith('.jpg')
                    assert file_name.startswith(('P', 'S', 'N', 'E', 'W'))

    def test_006(self):
        """Identify empty subset.

        This tests case consists of checking that is_empty() method
        works as expected on empty subset.
        """
        req = LucasRequest()
        req.bbox = (0, 0, 1, 1)
        lucasio = LucasIO()
        lucasio.download(req)

        assert lucasio.count() == 0
        assert lucasio.is_empty() is True

    def test_007(self):
        """Identify max feature property.

        This tests case consists of checking that MAX_FEATURES
        property.
        """
        req = LucasRequest()
        req.bbox = (3764067 ,2361825, 5030245, 3685331)
        lucasio = LucasIO()
        lucasio.download(req)

        assert lucasio.count() == int(lucasio.metadata['LUCAS_MAX_FEATURES'])



'''
Download ST_LUCAS samples based on built request.
'''

import os
import logging
import tempfile
import json
import requests
from pathlib import Path
from shutil import copy
from requests.exceptions import HTTPError, ConnectionError, InvalidSchema, InvalidURL, MissingSchema
from osgeo import gdal, ogr
import zipfile

gdal.UseExceptions()

from owslib.wfs import WebFeatureService
from owslib.util import ServiceException
from owslib import __version__ as owslib_version

from .logger import Logger
from .exceptions import LucasDownloadError, LucasDataError

__version__ = '1.1.1'

class LucasIO:
    """
    LUCAS features input / output class.

    :param str url: WFS endpoint
    :param str version: WFS version to be used
    """

    def __init__(self,
                 url='https://geoforall.fsv.cvut.cz/st_lucas',
                 version='1.1.0'):
        Logger.debug(f"Using owslib version {owslib_version}")
        self._wfs_url = url + "/geoserver/wfs"
        self._wfs_version = version

        self._mtd_url = url + "/st_lucas_metadata.json"
        self._request = None

        self._path = None # path to GPKG file

    @property
    def data(self):
        return self._path

    @data.setter
    def data(self, path):
        """Set data property from existing GPKG file.

        :param str path: path to existing GPKG file
        """
        try:
            ds = gdal.OpenEx(path, gdal.OF_VECTOR | gdal.OF_READONLY)
            driver = ds.GetDriver().ShortName
            if driver != "GPKG":
                raise LucasDataError(f"Unexpected input file: {driver}")
            del ds
        except RuntimeError as e:
            raise LucasDataError(f"Unable to open input file: {e}")

        self._path = path

    @property
    def metadata(self):
        """Get metadata.

        :return dict: metadata dictionary
        """
        try:
            ds = gdal.OpenEx(self._path, gdal.OF_VECTOR | gdal.OF_READONLY)
            md = ds.GetMetadata()
            del ds
        except RuntimeError as e:
            raise LucasDataError(f"Unable to get metadata: {e}")

        return md

    @staticmethod
    def __get_tempfile_name(extension):
        return os.path.join(tempfile.gettempdir(),
                            "st_lucas_{n}.{e}".format(
                                n=next(tempfile._get_candidate_names()),
                                e=extension)
                            )

    def download(self, request):
        """
        Download LUCAS features from dedicated WFS server based on
        specified request :class:`.request.LucasRequest`.

        :param LucasRequest: request
        """
        try:
            wfs = WebFeatureService(url=self._wfs_url, version=self._wfs_version)
        except (HTTPError, ConnectionError, InvalidSchema, InvalidURL, MissingSchema, AttributeError, UnicodeError) as e:
            raise LucasDownloadError(f"Cannot connect to server: {e}")


        Logger.debug(f"Connected to {self._wfs_url}")

        # collect getfeature() arguments
        args = {
            'srsname': "http://www.opengis.net/gml/srs/epsg.xml#3035"
        }
        args.update(request.build())
        Logger.debug(f"Request: {args}")

        try:
            response = wfs.getfeature(**args)
        except (ServiceException, AttributeError, TypeError) as e:
            raise LucasDownloadError(f"Unable to get features from WFS server: {e}")


        gml = response.read()
        Logger.info(
            "Download process successfuly finished. Size of downloaded data: {}kb".format(
                int(len(gml) / 1000)
            ))

        self._request = request

        self._path = self._load(gml)
        self._postprocessing(self._path)

    def _load(self, gml):
        """Load features from GML string and creates temporary GPKG file.

        :param str gml: GML string to be loaded

        :return str: path to temporary GPKG file
        """

        # 1. store GML string into temporary file
        path_gml = self.__get_tempfile_name("gml")
        Logger.debug(f"WFS GML path: {path_gml}")
        with open(path_gml, 'wb') as f:
            f.write(gml)

        # 2. convert GML to GPKG
        path_gpkg = self.__get_tempfile_name("gpkg")
        try:
            gdal.VectorTranslate(path_gpkg, path_gml, dstSRS="EPSG:3035")
        except RuntimeError as e:
            raise LucasDataError(f"Unable to translate into GPKG: {e}")

        return path_gpkg

    def _postprocessing(self, path):
        """Delete gml_id column. If data are space time aggregated, delete columns which aren't required

        :param str path: path to GPKG file
        """
        def _group(name):
            return {
                "LC_LU": "LAND COVER,LAND USE",
                "LC_LU_SO": "LAND COVER,LAND USE,SOIL",
                "FO": "FORESTRY",
                "CO": "COPERNICUS",
                "IN": "INSPIRE"
            }[name]

        try:
            ds = gdal.OpenEx(path, gdal.OF_VECTOR | gdal.OF_UPDATE)
            layer = ds.GetLayer()
            defn = layer.GetLayerDefn()
            layer.DeleteField(defn.GetFieldIndex('gml_id'))
            if self._request.st_aggregated and self._request.years is not None:
                driver = ogr.GetDriverByName("GPKG")
                gpkg = driver.Open(path)
                layer1 = gpkg.GetLayer()
                layer_definition = layer1.GetLayerDefn()
                for i in range(layer_definition.GetFieldCount()):
                    attr = layer_definition.GetFieldDefn(i).GetName()
                    try:
                        if int(attr[-4:]) not in self._request.years:
                            layer.DeleteField(defn.GetFieldIndex(attr))
                    except ValueError:
                        # some attributes are timeless (eg. point_id, ..., count_survey)
                        pass

            # read LUCAS JSON metadata from server
            try:
                r = requests.get(self._mtd_url)
                lucas_metadata = json.loads(r.content)
            except (HTTPError, json.decoder.JSONDecodeError) as e:
                raise LucasDataError(f"Postprocessing failed: {e}")

            # write metadata into GPKG
            # note:
            #  values are provided as strings due to GDAL < 3.3 limitation
            #  -> Dictionary must contain tuples of strings
            metadata = ({
                "LUCAS_TABLE": self._request.typename.split(":")[1],
                "LUCAS_ST": str(int(self._request.st_aggregated)),
                "LUCAS_CLIENT_VERSION": __version__,
                "LUCAS_DB_VERSION": str(lucas_metadata["version"]),
                "LUCAS_MAX_FEATURES": str(lucas_metadata["max_features"]),
            })
            if self._request.group is not None:
                metadata["LUCAS_GROUP"] = _group(self._request.group.upper())
            ds.SetMetadata(metadata)

            del ds
        except RuntimeError as e:
            raise LucasDataError(f"Postprocessing failed: {e}")

    def __check_data(self):
        """Check whether LUCAS features are downloaded.

        Raise LucasDownloadError of failure
        """
        if self._path is None:
            raise LucasDownloadError("No LUCAS features downloaded")

    def to_gml(self, epsg=3035):
        """Get downloaded LUCAS features as `OGC GML
        <https://www.ogc.org/standards/gml>`__ string.

        :param int epsg: target EPSG code

        :return str: GML string
        """
        self.__check_data()

        path_gml = self.__get_tempfile_name("gml")

        try:
            gdal.VectorTranslate(path_gml, self._path,
                                 dstSRS=f"EPSG:{epsg}",
                                 format="GML")
        except RuntimeError as e:
            raise LucasDataError(f"Unable to translate into GML: {e}")

        with open(path_gml) as fd:
            data = fd.read()

        return data

    def to_gpkg(self, output_path, epsg=3035):
        """Save downloaded LUCAS features into `OGC GeoPackage
        <https://www.ogc.org/standards/gml>`__ file.

        Raises LucasDataError on failure.

        :param str output_path: path to the output OGC GeoPackage file
        :param int epsg: target EPSG code
        """
        self.__check_data()

        out_path = Path(output_path)

        # Delete file if exists
        if out_path.exists():
            try:
                out_path.unlink()
            except PermissionError as e:
                raise LucasDataError(f"Unable to overwrite existing file: {e}")

        if epsg != 3035:
            try:
                gdal.VectorTranslate(output_path, self._path, dstSRS=f"EPSG:{epsg}")
            except RuntimeError as e:
                raise LucasDataError(f"Unable to translate into GPKG: {e}")
        else:
            # Copy file from temporary directory
            if Path(self._path).parent != out_path.parent:
                try:
                    copy(self._path, out_path.parent)
                except PermissionError as e:
                    raise LucasDataError(f"Permission denied: {e}")

            # Rename file
            path = Path(out_path.parent) / Path(self._path).name
            path.rename(out_path)

    def to_geopandas(self, epsg=3035):
        """Get downloaded LUCAS features as GeoPandas `GeoDataFrame
        <https://geopandas.org/docs/reference/api/geopandas.GeoDataFrame.html>`__
        structure.

        :param int epsg: target EPSG code

        :return GeoDataFrame:

        """
        from geopandas import read_file

        self.__check_data()

        if epsg != 3035:
            try:
                _path = self.__get_tempfile_name("gpkg")
                gdal.VectorTranslate(_path, self._path, dstSRS=f"EPSG:{epsg}")
            except RuntimeError as e:
                raise LucasDataError(f"Unable to translate into GPKG: {e}")
        else:
            _path = self._path

        return read_file(_path)

    def count(self):
        """Get number of downloaded LUCAS features.

        :return int: number of downloaded features
        """
        self.__check_data()

        try:
            ds = gdal.OpenEx(self._path, gdal.OF_VECTOR | gdal.OF_READONLY)
            layer = ds.GetLayer()
            nop = layer.GetFeatureCount()
            del ds
        except RuntimeError as e:
            raise LucasDataError(f"Postprocessing failed: {e}")

        return nop

    def is_empty(self):
        """Check whether downloaded LUCAS feature collection is empty.

        :return bool: True for empty collection otherwise False
        """
        return self.count() < 1

    def get_images(self, year, point_id, nuts0=None):
        """Get images of selected point and its surroundings from Eurostat FTP server.

        :param int year: year of the measurement
        :param int point_id: id of the LUCAS point
        :param str nuts0: NUTS0 code

        :return images: dictionary of images (URL)
        """
        if nuts0 is None:
            try:
                ds = gdal.OpenEx(self._path, gdal.OF_VECTOR | gdal.OF_READONLY)
                layer = ds.GetLayer(0)
                layer.SetAttributeFilter(f"point_id={point_id} AND survey_year={year}")
                num_points = layer.GetFeatureCount()
                if num_points != 1:
                    raise LucasDataError(f"Unexpected number of selected points: {num_points}")
                nuts0 = layer.GetNextFeature().GetFieldAsString("nuts0")
                del ds
            except RuntimeError as e:
                raise LucasDataError(f"Unable to get images: {e}")

        images = {}
        url = f'https://gisco-services.ec.europa.eu/lucas/photos/{str(year)}/{nuts0}/{str(point_id)[0:3]}/{str(point_id)[3:6]}/{str(point_id)}'
        for i in ("P", "S", "N", "E", "W"):
            images[i] = f"{url}{i}.jpg"

        return images

    def download_images(self, images, save_dir):
        """Save images of LUCAS point into zip file.

        :param dict images: dictionary with URLs of LUCAS point images received by get_images method
        :param str save_dir: directory path to save the downloaded images

        :return str: path to created zip file
        """
        if not Path(save_dir).exists():
            raise IOError(f"No such directory: {save_dir}")

        year = images["P"].split("/")[5]
        zip_name = images["P"].split("/")[-1].replace("P.jpg", f"_{year}.zip")
        path_to_zipfile = os.path.join(save_dir, zip_name)
        image_list = []
        for key, url in images.items():
            response = requests.get(url)
            if response.status_code == 200:
                image_list.append((key, response.content))

        if len(image_list) == 0:
            Logger.warning("None images available for given LUCAS point")
            return None

        with zipfile.ZipFile(path_to_zipfile, 'w') as zip_file:
            for key, image_data in image_list:
                zip_file.writestr(key + '.jpg', image_data)
        Logger.info(f"LUCAS photos downloaded to: {path_to_zipfile}")

        return path_to_zipfile



'''
Build request for quering LUCAS dataset.
'''
from owslib import __version__ as owslib_version
from owslib.etree import etree
from owslib.fes import Or, And, PropertyIsEqualTo, BBox, PropertyIsNull, Not

from .exceptions import LucasRequestError, LucasOWSLibError

def check_owslib():
    """Check supported OWSLib version.

    Raises LucasRequestError on unsupported OWSLib version.
    """
    owslib_min = 0.22
    owslib_max = 0.26
    owslib_ver = float('.'.join(owslib_version.split('.')[:2]))
    if owslib_ver > owslib_min and owslib_ver < owslib_max:
        raise LucasOWSLibError(
            f"Unsupported OWSLib version: {owslib_version}. "
            f"Supported versions: <={owslib_min}, >={owslib_max}"
        )

class LucasRequest:
    """Define a request.
    """
    bbox = None
    operator = None
    propertyname = None
    literal = None
    logical = None
    years = None
    countries = None
    aoi_polygon = None
    st_aggregated = False
    group = None

    gh_workspace = 'lucas'
    gh_typename = 'lucas_points'
    gh_typename_st = 'lucas_st_points'

    @property
    def typename(self):
        typename = self.gh_typename_st if self.st_aggregated else self.gh_typename
        if self.group is not None:
            typename += f'_{self.group}'.lower()
        return f'{self.gh_workspace}:{typename}'

    def __setattr__(self, name, value):
        if name not in (
                'bbox',
                'operator',
                'propertyname',
                'literal',
                'logical',
                'years',
                'countries',
                'aoi_polygon',
                'group',
                'st_aggregated'):
            raise LucasRequestError(f"{name} not allowed")

        super().__setattr__(name, value)

    def _check_args(self):
        """Check for required arguments.

        Raise LucasRequestError if requirements are not defined.
        """
        if self.bbox is None and \
           self.aoi_polygon is None and \
           self.countries is None and \
           self.propertyname is None:
            raise LucasRequestError(
                "spatial filter (bbox, countries, aoi_polygon) or attribute filter (propertyname) must be defined."
            )
        if self.bbox is not None and self.aoi_polygon is not None:
            raise LucasRequestError("bbox and aoi_polygon are mutually exclusive.")
        if self.bbox is not None and self.countries is not None:
            raise LucasRequestError("bbox and countries are mutually exclusive.")
        if self.aoi_polygon is not None and self.countries is not None:
            raise LucasRequestError("aoi_polygon and countries are mutually exclusive.")
        # check supported OWSLib version
        check_owslib()

    def __build_years_property(self, query=None):
        query_years = self.__build_years(query)
        if self.propertyname:
            return self.__build_property(
                self.propertyname.lower(), self.literal, self.operator, self.logical,
                query_years
            )
        return query_years

    def __build_years(self, query=None):
        filter_years = []
        if self.years is not None:
            for value in self.years:
                if self.st_aggregated:
                    filter_years.append(
                        Not([PropertyIsNull(propertyname=f'survey_date_{str(value)}')])
                    )
                else:
                    filter_years.append(
                        PropertyIsEqualTo(propertyname='survey_year',
                                          literal=str(value))
                    )

        if not filter_years:
            return query
        elif len(filter_years) > 1:
            if query:
                return And([
                    query,
                    Or(filter_years)
                ])
            else:
                return Or(filter_years)
        else:
            if query:
                return And([
                    query,
                    filter_years[0]
                ])
            else:
                return filter_years[0]

        return None

    def __build_property(self, propertyname, literal, operator, logical, query=None):
        # check requirements
        if propertyname is None:
            raise LucasRequestError(
                "Property name is not defined"
            )
        if literal is None:
            raise LucasRequestError(
                "Literal property is not defined"
            )

        # collect list of literals
        if isinstance(literal, str):
            literal = [literal]
        elif isinstance(literal, list):
            literal = literal
        elif isinstance(literal, int):
            literal = [str(literal)]
        else:
            raise LucasRequestError(
                "Literal property must be provided as a string or a list."
            )

        # check if logical operator is required
        if isinstance(literal, list) and len(literal) > 1 and logical is None:
            raise LucasRequestError(
                "Logical property is not defined."
            )

        # get list of property filters
        filter_list = []
        for value in literal:
            if propertyname in ('point_id', 'nuts0', 'survey_count', 'survey_maxdist') or self.st_aggregated is False:
                filter_list.append(
                    operator(propertyname=propertyname, literal=value)
                )
            else:
                st_filter = []
                for year in ['_2006', '_2009', '_2012', '_2015', '_2018']:
                    st_filter.append(
                        operator(propertyname=propertyname + year, literal=value)
                    )
                filter_list.append(Or(st_filter))


        # get combined property filter
        if len(literal) > 1:
            filter_property = logical(filter_list)
        else:
            filter_property = filter_list[0]

        if query:
            return And([query, filter_property])

        return filter_property

    def build(self):
        """
        Return a request arguments (bbox, filter) for the getfeature() method.

        :return dict:
        """
        self._check_args()

        req = {
            'typename': self.typename
        }

        #
        # filter by bbox
        #
        if self.bbox is not None:
            bbox_query = BBox(self.bbox)
            filter_ = self.__build_years_property(bbox_query)

            # TODO: workaround for owslib bug
            tree = filter_.toXML()
            namespaces = {'gml311': 'http://www.opengis.net/gml'}
            envelopes = tree.findall('.//gml311:Envelope', namespaces)
            for envelope in envelopes:
                envelope.set('srsName', 'http://www.opengis.net/gml/srs/epsg.xml#3035')

            filter_xml = etree.tostring(tree).decode("utf-8")
            # TODO maybe change this in XML way as well
            filter_xml = filter_xml.replace('ows:BoundingBox', 'geom')

            req.update({'filter': filter_xml})

        #
        # filter by countries
        #
        if self.countries is not None:
            countries_query = self.__build_property(
                'nuts0', self.countries, PropertyIsEqualTo, Or
            )
            filter_ = self.__build_years_property(countries_query)

            req.update({
                'filter': etree.tostring(filter_.toXML()).decode("utf-8")
            })

        #
        # filter by polygon
        #
        if self.aoi_polygon is not None:
            filter_ = self.__build_years_property()
            # merge with year filter if defined
            if filter_ is not None:
                filter_xml = etree.tostring(filter_.toXML()).decode("utf-8")
                filter_xml = '<ogc:And xmlns:ogc="http://www.opengis.net/ogc">' + filter_xml + self.aoi_polygon + '</ogc:And>'
            else:
                filter_xml = self.aoi_polygon

            req.update({'filter': filter_xml})

        if self.bbox is None and \
           self.countries is None and \
           self.aoi_polygon is None and \
           self.propertyname is not None:
            filter_ = self.__build_property(
                self.propertyname, self.literal, self.operator, self.logical
            )

            req.update({
                'filter': etree.tostring(filter_.toXML()).decode("utf-8")
            })

        return req



'''
Analyze downloaded LUCAS samples.
'''

import json
import sqlite3
import os
import csv

from osgeo import ogr

from st_lucas.exceptions import LucasDataError, LucasLoadError, LucasConfigError

class LucasClassAggregate:
    """Perform LC class aggregation.

    :param str gpkg_path: path to GPKG file created by :class:`.io.LucasIO.to_gpkg()`.
    """
    def __init__(self, gpkg_path, mappings=None, mappings_file=None):
        if mappings is not None and mappings_file is not None:
            raise LucasConfigError('Only one of the parameters "data" and '
                                   '"json_path" should be defined')
        if mappings is None and mappings_file is None:
            raise LucasConfigError('One of the parameters "data" and '
                                   '"json_path" should be defined')

        self._gpkg_path = gpkg_path
        self.mappings = mappings
        self.mappings_file = mappings_file

    def _load_classes(self, classes):
        """Load aggregation rules from JSON file.

        :param dict classes: defined aggregation rules

        :return dictionary: dictionary of original classes and names of aggregated classes
        """
        csv_lc1 = os.path.join(os.path.dirname(__file__), "lc1_codes.csv")
        with open(csv_lc1, newline='') as csv_f:
            layer_reader = csv.DictReader(csv_f, delimiter=";")
            # collect possible lc1 codes
            possible_codes = []
            for row in layer_reader:
                possible_codes.append(row["code"])

            try:
                values = list(classes.values())
                values_list = []
                for i in values:
                    for j in i:
                        values_list.append(j)
                        if j not in possible_codes:
                            raise LucasDataError(f"Code {j} is not Land Cover code!")

                if len(values_list) != len(set(values_list)):
                    raise LucasDataError("Some code is used repeatedly!")

            except ValueError as e:
                raise LucasDataError(f"Invalid json file: {e}")

        return classes

    def apply(self):
        """Apply aggregation rules on GPKG file

        :param dict data: defined aggregation rules
        """
        if self.mappings is not None:
            self._apply_from_data()
        else:
            self._apply_from_file()

    def _apply_from_file(self):
        """Apply aggregation rules defined in JSON file on GPKG file

        :param str json_path: path to JSON file with defined aggregation rules
        """
        try:
            with open(self.mappings_file) as json_file:
                self.mappings = json.load(json_file)
                self._apply_from_data()
        except FileNotFoundError as e:
            raise LucasLoadError(f"Invalid json file path: {e}")

    def _apply_from_data(self):
        """Apply aggregation rules on GPKG file

        :param dict data: defined aggregation rules
        """
        driver = ogr.GetDriverByName("GPKG")
        if os.path.exists(self._gpkg_path):
            gpkg = driver.Open(self._gpkg_path)
            layer = gpkg.GetLayer()
            layer_name = layer.GetName()

            if layer_name[6:8] == "st":
                columns_h = []
                columns_a = []
                layer_definition = layer.GetLayerDefn()
                for i in range(layer_definition.GetFieldCount()):
                    attr = layer_definition.GetFieldDefn(i).GetName()
                    if attr in ["lc1_h_2006", "lc1_h_2009", "lc1_h_2012", "lc1_h_2015", "lc1_h_2018"]:
                        columns_h.append(attr)
                        columns_a.append(attr.replace("h", "a"))
                if not columns_h:
                    raise LucasDataError(f"There is no lc1_h column in gpkg file!")
            else:
                columns_h = ["lc1_h"]
                columns_a = ["lc1_a"]


            classes = self._load_classes(self.mappings)
            with sqlite3.connect(self._gpkg_path) as con:
                con.enable_load_extension(True)
                cur = con.cursor()
                cur.execute('SELECT load_extension("mod_spatialite");')
                for h_column, new_column in zip(columns_h, columns_a):
                    try:
                        cur.execute(
                            f"ALTER TABLE {layer_name} ADD COLUMN {new_column} "
                            f"TEXT")
                    except sqlite3.OperationalError:
                        sql_query = f"UPDATE {layer_name} SET {new_column} = " \
                                    f"null"
                        cur.execute(sql_query)
                        print(f"Column {new_column} already exists in the table "
                              f"{layer_name} - it will be rewritten with the new "
                              f"data")
                    try:
                        cur.execute(f"CREATE INDEX IF NOT EXISTS {h_column}_idx ON {layer_name}({h_column})")
                        for key in classes:
                            q_marks = "?" * len(classes[key])
                            sql_query = f"UPDATE {layer_name} SET {new_column} =? WHERE {h_column} IN ({','.join(q_marks)})"
                            val = tuple([key] + classes[key])
                            cur.execute(sql_query, val)
                    except sqlite3.OperationalError as e:
                        raise LucasDataError(f"Not possible to aggregate: {e}")
            con.close()

        else:
            raise LucasLoadError("GPKG file doesn't exist")


class LucasClassTranslate:
    """Perform LC class translation.

    :param str gpkg_path: path to GPKG file created by :class:`.io.LucasIO.to_gpkg()`.
    """
    def __init__(self, gpkg_path, csvpath=None):
        if csvpath is None:
            csvpath = os.path.join(os.path.split(__file__)[0],
                                   'nomenclature_translation',
                                   'LUCAS_unique_lc1_lu1_combinations_ALL.csv')

        self.gpkg_path = gpkg_path
        self.csv_path = csvpath
        self.source_col = 'CLC2'

        if not os.path.exists(self.gpkg_path):
            raise LucasLoadError(f"GPKG file {self.gpkg_path} not found")
        if not os.path.exists(self.csv_path):
            raise LucasLoadError(f"CSV file {self.csv_path} not found")

    def get_translations(self):
        """Get supported translations.

        :return list: list of translation tables
        """
        with open(self.csv_path) as trans_data:
            return trans_data.readline().split(',')

    def set_translations(self, source_col):
        """Get supported translations.

        :return list: list of translation tables
        """
        self.source_col = source_col

    def apply(self):
        """Apply translation rules on GPKG file.
        """
        driver = ogr.GetDriverByName("GPKG")

        gpkg = driver.Open(self.gpkg_path)
        layer = gpkg.GetLayer()
        layer_name = layer.GetName()

        with open(self.csv_path) as trans_data:
            header = trans_data.readline()
            header_list = header.split(',')

            #new_col = 'clc3'
            lc1_col = 'lc1'
            lu1_col = 'lu1'

            if self.source_col not in header_list:
                raise LucasConfigError(
                    f'Column {self.source_col} not found in the header. '
                    f'Only columns {header} found')

            new_col_ind = header_list.index(self.source_col)
            lc1_ind = header_list.index('LC1')
            lu1_ind = header_list.index('LU1')
            repre_ind = header_list.index('Representativeness')

            new_col = self.source_col.lower()

            with sqlite3.connect(self.gpkg_path) as con:
                con.enable_load_extension(True)
                cur = con.cursor()
                cur.execute('SELECT load_extension("mod_spatialite");')
                try:
                    cur.execute(
                        f"ALTER TABLE {layer_name} ADD COLUMN {new_col} TEXT")
                except sqlite3.OperationalError:
                    sql_query = f"UPDATE {layer_name} SET {new_col} = null"
                    cur.execute(sql_query)
                    print(f"Column {new_col} already exists in the table "
                          f"{layer_name} - it will be rewritten with the new "
                          f"data")

                sql_query = f"UPDATE {layer_name} SET {new_col} = ? WHERE " \
                            f"{lc1_col} = ? AND {lu1_col} = ?"

                for line in trans_data.readlines():
                    line_split = line.split(',')
                    if line_split[repre_ind] == '1':
                        cur.execute(sql_query,
                                    (line_split[new_col_ind],
                                     line_split[lc1_ind],
                                     line_split[lu1_ind]))



'''
LUCAS-related logging.
'''

import os
import logging
import logging.config

class LucasRequestLogger(logging.getLoggerClass()):
    pass

def logger():
    """Return a logger.
    """
    logging.config.fileConfig(
        os.path.join(os.path.dirname(__file__), 'logging.conf')
    )

    logging.setLoggerClass(LucasRequestLogger)
    logger = logging.getLogger('LUCAS')

    return logger

Logger = logger()



import setuptools
from st_lucas import __version__
from pathlib import Path

root_dir = Path(__file__).parent

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='st_lucas',
    version=__version__,
    description='ST_LUCAS Python package',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/geoharmonizer_inea/st_lucas/st_lucas-python-package',
    packages=setuptools.find_packages(),
    package_data={
        setuptools.find_packages()[0]: [
            "logging.conf",
            "analyze/lc1_codes.csv",
            "analyze/nomenclature_translation/*"
        ]
    },
    scripts=[],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
    install_requires=[
        'geopandas>=0.8',
        'OWSLib>=0.26',
        'GDAL>=3.5',
    ],
)



'''
LUCAS-related exceptions.
'''

class LucasRequestError(Exception):
    """Building request failed"""
    pass

class LucasDownloadError(Exception):
    """Data download failed"""
    pass

class LucasDataError(Exception):
    """Data processing failed"""
    pass

class LucasLoadError(Exception):
    """File open failed"""
    pass

class LucasConfigError(Exception):
    """File open failed"""
    pass

class LucasOWSLibError(Exception):
    """Unsupported OWSLib version"""
    pass
